#pragma once 
#ifndef _XMLPROPBAG_H_
#define _XMLPROPBAG_H_
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
//
// XMLPropertyBag example - Copyright 1999, DevelopMentor
//
// Sept. 4, 1999 - Don Box
//
// This file implements three functions:
//
//    GetPropertyBagFromXML - Builds a property bag on top of an existing XML element
//    SaveToXML - Saves an IPersistPropertyBag-based object to an XML stream
//    LoadFromXML - Loads an IPersistPropertyBag-based object from an XML stream
//
//  This implementation supports everything you can put into a variant EXCEPT arrays.
//  FWIW, VB doesn't support writing arrays or UDTs to a property bag. This code
//  supports UDTs but not arrays (for now).
// 
//  This code also assumes that object references are unique. If you try to store
//  the same object reference twice, I serialize the object twice.
//
//

#define _WIN32_DCOM
#include <windows.h>
#include <ocidl.h>
#include <atlbase.h>
#include <atlcom.h>
#include <atldb.h>
#include <msXml.h>


HRESULT GetPropertyBagFromXML(IXMLDOMDocument *pDocument, IXMLDOMElement *pElement, IPropertyBag **ppPropBag);
HRESULT SaveToXML(IXMLDOMDocument *pDocument, IXMLDOMElement *pElement, IPersistPropertyBag *pObj, BOOL bClearDirty, BOOL bSaveAllProperties);
HRESULT LoadFromXML(IXMLDOMDocument *pDocument, IXMLDOMElement *pElement, IErrorLog *pLog, IPersistPropertyBag **ppObj);


class XMLPropertyBag : public CComObjectRootEx<CComMultiThreadModelNoCS>, 
                       public IPropertyBag
{
public:
    BEGIN_COM_MAP(XMLPropertyBag)
        COM_INTERFACE_ENTRY(IPropertyBag)
    END_COM_MAP()
    CComPtr<IXMLDOMDocument> m_doc;
    CComPtr<IXMLDOMElement>  m_elem;


    STDMETHODIMP Read(LPCOLESTR pwszPropName, VARIANT *pVar, IErrorLog *pErrorLog)
    {
        CComPtr<IXMLDOMNode> node;
        HRESULT hr = m_elem->selectSingleNode(CComBSTR(pwszPropName), &node);
        if (FAILED(hr)) return hr;
        if (hr == S_FALSE) 
        {
            pVar->vt = VT_EMPTY;
            return S_OK;
        }

        CComPtr<IXMLDOMElement> pElem;
        hr = node->QueryInterface(&pElem);
        if (FAILED(hr)) return hr;

        CComVariant varVT;
        hr = pElem->getAttribute(CComBSTR(L"vt"), &varVT);
        if (FAILED(hr)) return hr;
        
        hr = varVT.ChangeType(VT_I2);
        if (FAILED(hr)) return hr;

        VARTYPE vt = varVT.iVal;

        if (vt == VT_RECORD)
        {
            CComVariant var;
            hr = pElem->getAttribute(CComBSTR(L"uuid"), &var);
            if (FAILED(hr)) return hr;
            hr = var.ChangeType(VT_BSTR);
            if (FAILED(hr)) return hr;
            GUID uuid;
            hr = CLSIDFromString(var.bstrVal, &uuid);
            if (FAILED(hr)) return hr;
            var.Clear();

            hr = pElem->getAttribute(CComBSTR(L"libid"), &var);
            if (FAILED(hr)) return hr;
            hr = var.ChangeType(VT_BSTR);
            if (FAILED(hr)) return hr;
            GUID libid;
            hr = CLSIDFromString(var.bstrVal, &libid);
            if (FAILED(hr)) return hr;
            var.Clear();

            hr = pElem->getAttribute(CComBSTR(L"major"), &var);
            if (FAILED(hr)) return hr;
            hr = var.ChangeType(VT_I2);
            if (FAILED(hr)) return hr;
            WORD wMajor = var.iVal;
            var.Clear();

            hr = pElem->getAttribute(CComBSTR(L"minor"), &var);
            if (FAILED(hr)) return hr;
            hr = var.ChangeType(VT_I2);
            if (FAILED(hr)) return hr;
            WORD wMinor = var.iVal;
            var.Clear();

            hr = pElem->getAttribute(CComBSTR(L"lcid"), &var);
            if (FAILED(hr)) return hr;
            hr = var.ChangeType(VT_UI4);
            if (FAILED(hr)) return hr;
            LCID lcid = var.ulVal;
            var.Clear();

            CComPtr<IRecordInfo> pRecInfo;
            hr = GetRecordInfoFromGuids(libid, wMajor, wMinor, lcid, uuid, &pRecInfo);
            if (FAILED(hr)) return hr;

            void *pvRecord = pRecInfo->RecordCreate();

            ULONG cFields;
            hr = pRecInfo->GetFieldNames(&cFields, 0);
            if (FAILED(hr)) return hr;

            BSTR *prgNames = new BSTR[cFields];
            CAutoMemRelease<BSTR> autoFree(prgNames);
            hr = pRecInfo->GetFieldNames(&cFields, prgNames);
            if (FAILED(hr)) return hr;

            CComPtr<IPropertyBag> pPropBag;
            hr = GetPropertyBagFromXML(m_doc.p, pElem.p, &pPropBag);
            for (ULONG i = 0; i < cFields; i++)
            {
                if (SUCCEEDED(hr))
                {
                    CComVariant value;
                    hr = pPropBag->Read(prgNames[i], &value, pErrorLog);
                    if (SUCCEEDED(hr))
                        hr = pRecInfo->PutField(INVOKE_PROPERTYPUT, pvRecord, prgNames[i], &value);\
                }
                SysFreeString(prgNames[i]);
            }
            if (FAILED(hr)) return hr;
            pVar->vt = VT_RECORD;
            pVar->pRecInfo = pRecInfo.Detach();
            pVar->pvRecord = pvRecord;

        }
        else if (vt == VT_UNKNOWN || vt == VT_DISPATCH)
        {
            CComPtr<IPersistPropertyBag> pppb;
            hr = LoadFromXML(m_doc.p, pElem.p, pErrorLog, &pppb);
            if (FAILED(hr)) return hr;

            pVar->vt = vt;
            if (hr == S_OK)
                return pppb->QueryInterface((vt == VT_UNKNOWN ? IID_IUnknown : IID_IDispatch), (void**)&pVar->punkVal);
            else
            {
                pVar->punkVal = 0;
                return S_FALSE;
            }
        }
        else if (vt == VT_BSTR)
        {
            hr = pElem->get_text(&pVar->bstrVal);
            if (FAILED(hr)) return hr;
            pVar->vt = vt;
        }
        else
        {
            CComVariant textValue;
            hr = pElem->get_text(&textValue.bstrVal);
            if (FAILED(hr)) return hr;
            textValue.vt = VT_BSTR;
            return VariantChangeType(pVar, &textValue, 0, vt);
        }


        return S_OK;
    }

    STDMETHODIMP Write(LPCOLESTR pwszPropName, VARIANT *pVar)
    {
        VARTYPE vt = pVar->vt;
        if (vt & VT_BYREF)
            return E_NOTIMPL;


        CComPtr<IXMLDOMElement> pElem;
        HRESULT hr = m_doc->createElement(CComBSTR(pwszPropName), &pElem);
        if (FAILED(hr)) return hr;

        hr = pElem->setAttribute(CComBSTR(L"vt"), CComVariant(vt));
        if (FAILED(hr)) return hr;

        if (vt & VT_ARRAY)
        {
            return E_NOTIMPL;
        }
        else if (vt == VT_RECORD)
        {
            CComPtr<IRecordInfo> pRecInfo = pVar->pRecInfo;
            void *pvRecord = pVar->pvRecord;

            GUID guid;
            hr = pRecInfo->GetGuid(&guid);
            if (FAILED(hr)) return hr;

            CComBSTR bstr(guid);
            hr = pElem->setAttribute(CComBSTR(L"uuid"), CComVariant(bstr));
            if (FAILED(hr)) return hr;


            CComPtr<ITypeInfo> pTypeInfo;
            hr = pRecInfo->GetTypeInfo(&pTypeInfo);
            if (FAILED(hr)) return hr;

            CComPtr<ITypeLib> pTypeLib;
            UINT index;
            hr  = pTypeInfo->GetContainingTypeLib(&pTypeLib, &index);
            if (FAILED(hr)) return hr;

            TLIBATTR *pattr = 0;
            hr = pTypeLib->GetLibAttr(&pattr);
            if (FAILED(hr)) return hr;

            hr = pElem->setAttribute(CComBSTR(L"libid"), CComVariant(CComBSTR(pattr->guid)));
            if (FAILED(hr)) return hr;
            hr = pElem->setAttribute(CComBSTR(L"major"), CComVariant(pattr->wMajorVerNum));
            if (FAILED(hr)) return hr;
            hr = pElem->setAttribute(CComBSTR(L"minor"), CComVariant(pattr->wMinorVerNum));
            if (FAILED(hr)) return hr;
            hr = pElem->setAttribute(CComBSTR(L"lcid"), CComVariant(long(pattr->lcid)));
            if (FAILED(hr)) return hr;

            pTypeLib->ReleaseTLibAttr(pattr);

            ULONG cFields;
            hr = pRecInfo->GetFieldNames(&cFields, 0);
            if (FAILED(hr)) return hr;

            BSTR *prgNames = new BSTR[cFields];
            CAutoMemRelease<BSTR> autoFree(prgNames);
            hr = pRecInfo->GetFieldNames(&cFields, prgNames);
            if (FAILED(hr)) return hr;

            CComPtr<IPropertyBag> pPropBag;
            hr = GetPropertyBagFromXML(m_doc.p, pElem.p, &pPropBag);
            for (ULONG i = 0; i < cFields; i++)
            {
                if (SUCCEEDED(hr))
                {
                    CComVariant value;
                    hr = pRecInfo->GetField(pvRecord, prgNames[i], &value);
                    if (SUCCEEDED(hr))
                        hr = pPropBag->Write(prgNames[i], &value);
                }
                SysFreeString(prgNames[i]);
            }
            if (FAILED(hr)) return hr;

        }
        else if (vt == VT_UNKNOWN || vt == VT_DISPATCH)
        {
            CComPtr<IPersistPropertyBag> pppb;
            if (pVar->punkVal)
            {
                hr = pVar->punkVal->QueryInterface(&pppb);
                if (FAILED(hr)) return hr;
            }

            hr = SaveToXML(m_doc.p, pElem.p, pppb.p, TRUE, TRUE);
            if (FAILED(hr)) return hr;
        }
        else if (vt == VT_BSTR)
        {
            hr = pElem->put_text(pVar->bstrVal);
            if (FAILED(hr)) return hr;
        }
        else
        {
            CComVariant value;
            hr = VariantChangeType(&value, pVar, 0, VT_BSTR);
            if (FAILED(hr)) return hr;

            hr = pElem->put_text(value.bstrVal);
            if (FAILED(hr)) return hr;

        }

        return m_elem->appendChild(pElem.p, 0);
    }
};


inline HRESULT GetPropertyBagFromXML(IXMLDOMDocument *pDocument, IXMLDOMElement *pElement, IPropertyBag **ppPropBag)
{
    CComObject<XMLPropertyBag> *pPropBag = 0;
    HRESULT hr = pPropBag->CreateInstance(&pPropBag);
    if (FAILED(hr)) return hr;

    pPropBag->m_doc = pDocument;
    pPropBag->m_elem = pElement;
    return pPropBag->QueryInterface(ppPropBag);
}

inline HRESULT SaveToXML(IXMLDOMDocument *pDocument, IXMLDOMElement *pElement, IPersistPropertyBag *pObj, BOOL bClearDirty, BOOL bSaveAllProperties)
{
    CComPtr<IPropertyBag> pPropBag;
    HRESULT hr = GetPropertyBagFromXML(pDocument, pElement, &pPropBag);
    if (FAILED(hr)) return hr;
    CLSID clsid = { 0 };
    if (pObj)
    {
        hr = pObj->GetClassID(&clsid);
        if (FAILED(hr)) return hr;
    }

    WCHAR wszGUID[40];
    StringFromGUID2(clsid, wszGUID, sizeof(wszGUID));

    hr = pElement->setAttribute(CComBSTR(L"clsid"), CComVariant(wszGUID));
    if (FAILED(hr)) return hr;

    if (pObj)
        return pObj->Save(pPropBag, bClearDirty, bSaveAllProperties);
    else
        return S_OK;
}

inline HRESULT LoadFromXML(IXMLDOMDocument *pDocument, IXMLDOMElement *pElement, IErrorLog *pLog, IPersistPropertyBag **ppObj)
{
    CComVariant varCLSID;
    HRESULT hr = pElement->getAttribute(CComBSTR(L"clsid"), &varCLSID);
    if (FAILED(hr) || varCLSID.vt != VT_BSTR) return hr;

    CLSID clsid;
    hr = CLSIDFromString(varCLSID.bstrVal, &clsid);
    if (FAILED(hr)) return hr;

    if (clsid == GUID_NULL)
    {
        *ppObj = 0;
        return S_FALSE;
    }

    CComPtr<IPersistPropertyBag> pObj;
    hr = pObj.CoCreateInstance(clsid);
    if (FAILED(hr)) return hr;


    CComPtr<IPropertyBag> pPropBag;
    hr = GetPropertyBagFromXML(pDocument, pElement, &pPropBag);
    if (FAILED(hr)) return hr;

    hr = pObj->Load(pPropBag, pLog);
    if (FAILED(hr)) return hr;

    *ppObj = pObj.Detach();

    return S_OK;

}

#endif